import urwid


may_use_fancy_formats = not hasattr(urwid.escape, "_fg_attr_xterm")


def add_setting(color, setting):
    if may_use_fancy_formats:
        return f"{color}, {setting}"
    return color


inheritance_overrides = {}


def link(child: str, parent: str):
    inheritance_overrides[child] = parent


def reset_inheritance_overrides():
    inheritance_overrides.clear()



from __future__ import annotations


COMMAND = {"zsh": "{_command_names -e}"}
PREAMBLE = {
    "zsh": """\
_script_args() {
  # pudb -m <TAB>
  if (($words[(I)-m] == $#words - 1)); then
    _python_modules
  # pudb -m XXX <TAB>
  elif (($words[(I)-m])); then
    _files
  # pudb <TAB>
  else
    _arguments -S -s '(-)1:script_args:_files -g "*.py"' '*: :_files'
  fi
}
""",
}
SCRIPT_ARGS = {"zsh": "_script_args"}


def get_argparse_parser():
    import argparse
    import os
    import sys
    try:
        import shtab
    except ImportError:
        from . import _shtab as shtab

    from pudb import VERSION

    version_info = "%(prog)s v" + VERSION

    if sys.argv[1:] == ["-v"]:
        print(version_info % {"prog": "pudb"})
        sys.exit(os.EX_OK)

    parser = argparse.ArgumentParser(
        "pudb",
        usage="%(prog)s [options] [-m] SCRIPT-OR-MODULE-TO-RUN [SCRIPT_ARGS]",
        epilog=version_info
    )
    shtab.add_argument_to(parser, preamble=PREAMBLE)
    # dest="_continue_at_start" needed as "continue" is a python keyword
    parser.add_argument(
        "-c", "--continue",
        action="store_true",
        dest="_continue_at_start",
        help="Let the script run until an exception occurs or a breakpoint is hit",
    )
    parser.add_argument("-s", "--steal-output", action="store_true")

    # note: we're implementing -m as a boolean flag, mimicking pdb's behavior,
    # and makes it possible without much fuss to support cases like:
    #    python -m pudb -m http.server -h
    # where the -h will be passed to the http.server module
    parser.add_argument("-m", "--module", action="store_true",
                        help="Debug as module or package instead of as a script")

    parser.add_argument("-le", "--log-errors", nargs=1, metavar="FILE",
                        help="Log internal errors to the given file"
                        ).complete = shtab.FILE
    parser.add_argument("--pre-run", metavar="COMMAND",
                        help="Run command before each program run",
                        default="").complete = COMMAND
    parser.add_argument("--version", action="version", version=version_info)
    parser.add_argument("script_args", nargs=argparse.REMAINDER,
                        help="Arguments to pass to script or module"
                        ).complete = SCRIPT_ARGS
    return parser


def main(**kwargs):
    import sys

    parser = get_argparse_parser()

    options = parser.parse_args()
    args = options.script_args

    if options.log_errors:
        from pudb.lowlevel import setlogfile
        setlogfile(options.log_errors[0])

    options_kwargs = {
        "pre_run": options.pre_run,
        "steal_output": options.steal_output,
        "_continue_at_start": options._continue_at_start,
    }

    if len(args) < 1:
        parser.print_help()
        sys.exit(2)

    mainpyfile = args[0]
    sys.argv = args

    if options.module:
        from pudb import runmodule
        runmodule(mainpyfile, **options_kwargs)
    else:
        from os.path import exists
        if not exists(mainpyfile):
            print(f"Error: {mainpyfile} does not exist", file=sys.stderr)
            sys.exit(1)

        from pudb import runscript
        runscript(mainpyfile, **options_kwargs)


if __name__ == "__main__":
    main()



from pudb.themes.utils import add_setting, link


# Give the colors some comprehensible names
black = "h235"
blacker = "h233"
dark_cyan = "h24"
dark_gray = "h241"
dark_green = "h22"
dark_red = "h88"
dark_teal = "h23"
light_blue = "h111"
light_cyan = "h80"
light_gray = "h252"
light_green = "h113"
light_red = "h160"
medium_gray = "h246"
salmon = "h223"
orange = "h173"
white = "h255"
yellow = "h192"

link("focused breakpoint", "focused selectable")
link("current breakpoint", "current frame name")
link("focused current breakpoint", "focused current frame name")

palette_dict = {
    # {{{ base styles
    "background": (black, light_gray),
    "selectable": (white, blacker),
    "focused selectable": (yellow, dark_cyan),
    "hotkey": (add_setting(black, "underline"), light_gray),
    "highlighted": (white, dark_green),
    # }}}
    # {{{ general ui
    "focused sidebar": (dark_cyan, light_gray),
    "group head": (add_setting(dark_cyan, "bold"), light_gray),
    "dialog title": (add_setting(light_gray, "bold"), black),
    "warning": (add_setting(white, "bold"), dark_red),
    "fixed value": (add_setting(white, "bold"), dark_gray),
    "button": (add_setting(white, "bold"), black),
    "focused button": (add_setting(yellow, "bold"), dark_cyan),
    # }}}
    # {{{ source view
    "line number": (dark_gray, black),
    "current line marker": (add_setting(yellow, "bold"), black),
    "breakpoint marker": (add_setting(light_red, "bold"), black),
    "source": (white, black),
    "breakpoint source": (add_setting(white, "bold"), dark_red),
    "current source": (add_setting(light_gray, "bold"), dark_teal),
    # }}}
    # {{{ sidebar
    "sidebar two": (light_blue, blacker),
    "focused sidebar two": (light_gray, dark_cyan),
    "sidebar three": (medium_gray, blacker),
    "focused sidebar three": (salmon, dark_cyan),
    # }}}
    # {{{ variables view
    "highlighted var label": (light_gray, dark_green),
    "return label": (light_green, blacker),
    "focused return label": (add_setting(light_gray, "bold"), dark_cyan),
    # }}}
    # {{{ stack
    "current frame name": (yellow, blacker),
    "focused current frame name": (add_setting(yellow, "bold"), dark_cyan),
    # }}}
    # {{{ shell
    "command line prompt": (add_setting(yellow, "bold"), black),
    "command line output": (light_cyan, black),
    "command line error": (light_red, black),
    # }}}
    # {{{ Code syntax
    "comment": (medium_gray, black),
    "exception": (orange, black),
    "function": (yellow, black),
    "keyword": (light_blue, black),
    "literal": (orange, black),
    "operator": (yellow, black),
    "pseudo": (medium_gray, black),
    "punctuation": (salmon, black),
    "string": (light_green, black),
    # }}}
}

# vim: foldmethod=marker



from pudb.themes.utils import add_setting, link


link("current breakpoint", "current frame name")
link("focused current breakpoint", "focused current frame name")

palette_dict = {
    # {{{ base styles
    "background": ("black", "light gray"),
    "selectable": ("white", "black"),
    "focused selectable": ("white", "dark gray"),
    "highlighted": ("black", "dark green"),
    "hotkey": (add_setting("black", "underline"), "light gray"),
    # }}}
    # {{{ general ui
    "input": ("white", "black"),
    "button": (add_setting("white", "bold"), "black"),
    "focused button": (add_setting("white", "bold"), "dark gray"),
    "focused sidebar": ("dark blue", "light gray"),
    "warning": (add_setting("white", "bold"), "dark red"),
    "group head": (add_setting("black", "bold"), "light gray"),
    "dialog title": (add_setting("white", "bold"), "black"),
    # }}}
    # {{{ source view
    "current source": ("black", "dark cyan"),
    "breakpoint source": (add_setting("white", "bold"), "dark red"),
    "line number": ("dark gray", "black"),
    "current line marker": (add_setting("dark cyan", "bold"), "black"),
    "breakpoint marker": (add_setting("dark red", "bold"), "black"),
    # }}}
    # {{{ sidebar
    "sidebar two": ("light cyan", "black"),
    "focused sidebar two": ("light cyan", "dark gray"),
    "sidebar three": ("light magenta", "black"),
    "focused sidebar three": ("light magenta", "dark gray"),
    # }}}
    # {{{ variables view
    "return label": ("light green", "black"),
    "focused return label": ("light green", "dark gray"),
    # }}}
    # {{{ stack
    "current frame name": ("light green", "black"),
    "focused current frame name": ("light green", "dark gray"),
    # }}}
    # {{{ shell
    "command line prompt": (add_setting("yellow", "bold"), "black"),
    "command line output": ("light cyan", "black"),
    "command line error": ("yellow", "black"),
    "focused command line output": ("light cyan", "dark gray"),
    "focused command line error": (add_setting("yellow", "bold"), "dark gray"),
    # }}}
    # {{{ Code syntax
    "literal":   ("light magenta", "black"),
    "builtin":   ("light cyan", "black"),
    "exception": ("light cyan", "black"),
    "keyword2":  ("light cyan", "black"),
    "function":  ("light green", "black"),
    "class":     (add_setting("light green", "underline"), "black"),
    "keyword":   ("light red", "black"),
    "operator":  ("light red", "black"),
    "comment":   ("dark gray", "black"),
    "docstring": ("dark gray", "black"),
    "argument":  ("brown", "black"),
    "pseudo":    ("brown", "black"),
    "string":    ("yellow", "black"),
    # }}}
}

# vim: foldmethod=marker



from pudb.themes.utils import add_setting, link


# Give the colors some comprehensible names
black = "h236"
blacker = "h234"
dark_gray = "h240"
dark_green = "h28"
dark_red = "h124"
dark_teal = "h30"
dark_magenta = "h141"
light_blue = "h111"
light_cyan = "h51"
light_gray = "h252"
light_green = "h155"
light_red = "h160"
light_magenta = "h198"
medium_gray = "h243"
orange = "h208"
white = "h255"
yellow = "h228"

link("current breakpoint", "current frame name")
link("focused current breakpoint", "focused current frame name")

palette_dict = {
    # {{{ base styles
    "background": (black, light_gray),
    "selectable": (white, blacker),
    "focused selectable": (white, dark_gray),
    "highlighted": (white, dark_green),
    "hotkey": (add_setting(black, "underline"), light_gray),
    # }}}
    # {{{ general ui
    "input": (white, black),
    "button": (add_setting(white, "bold"), black),
    "focused button": (add_setting(white, "bold"), dark_gray),
    "focused sidebar": (dark_teal, light_gray),
    "warning": (add_setting(white, "bold"), dark_red),
    "group head": (add_setting(black, "bold"), light_gray),
    "dialog title": (add_setting(white, "bold"), blacker),
    # }}}
    # {{{ source view
    "source": (white, black),
    "current source": (add_setting(light_gray, "bold"), dark_teal),
    "breakpoint source": (add_setting(white, "bold"), dark_red),
    "line number": (dark_gray, black),
    "current line marker": (add_setting(light_cyan, "bold"), black),
    "breakpoint marker": (add_setting(light_red, "bold"), black),
    # }}}
    # {{{ sidebar
    "sidebar two": (light_cyan, blacker),
    "focused sidebar two": (light_cyan, dark_gray),
    "sidebar three": (dark_magenta, blacker),
    "focused sidebar three": (dark_magenta, dark_gray),
    # }}}
    # {{{ variables view
    "highlighted var label": (light_gray, dark_green),
    "return label": (light_green, blacker),
    "focused return label": (light_green, dark_gray),
    # }}}
    # {{{ stack
    "current frame name": (light_green, blacker),
    "focused current frame name": (light_green, dark_gray),
    # }}}
    # {{{ shell
    "command line prompt": (add_setting(yellow, "bold"), black),
    "command line output": (light_cyan, black),
    "command line error": (orange, black),
    "focused command line output": (light_cyan, dark_gray),
    "focused command line error": (add_setting(orange, "bold"), dark_gray),
    # }}}
    # {{{ Code syntax
    "literal":     (dark_magenta, black),
    "builtin":     (light_cyan, black),
    "exception":   (light_cyan, black),
    "keyword2":    (light_cyan, black),
    "function":    (light_green, black),
    "class":       (add_setting(light_green, "underline"), black),
    "keyword":     (light_magenta, black),
    "operator":    (light_magenta, black),
    "comment":     (medium_gray, black),
    "docstring":   (medium_gray, black),
    "argument":    (orange, black),
    "pseudo":      (orange, black),
    "string":      (yellow, black),
    # }}}
}

# vim: foldmethod=marker



from pudb.themes.utils import add_setting, link


# ------------------------------------------------------------------------------
# Colors are approximations of https://www.nordtheme.com/docs/colors-and-palettes
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
# Polar Night is made up of four darker colors (nord 0, nord1, nord2, and nord3)
# that are commonly used for base elements like backgrounds or text color in
# bright ambiance designs.
# ------------------------------------------------------------------------------

# nord0 is the origin color or the Polar Night palette.
nord0 = "h236"

# nord1 is a brighter shade color based on nord0.
nord1 = "h238"

# nord2 is an even more brighter shade color of nord0.
nord2 = "h239"

# nord3 is the brightest shade color based on nord0.
nord3 = "h243"

# ------------------------------------------------------------------------------
# Snow Storm is made up of three bright colors that are commonly used for text
# colors or base UI elements in bright ambiance designs.
# ------------------------------------------------------------------------------

# nord4 is the origin color or the Snow Storm palette.
nord4 = "h253"

# nord5 is a brighter shade color of nord4.
nord5 = "h254"

# nord6 is the brightest shade color based on nord4.
nord6 = "h255"

# ------------------------------------------------------------------------------
# Frost can be described as the heart palette of Nord, a group of four bluish
# colors that are commonly used for primary UI component and text highlighting
# and essential code syntax elements.
# ------------------------------------------------------------------------------

# nord7 is a calm and highly contrasted color reminiscent of
# frozen polar water.
nord7 = "h109"

# nord8 is a bright and shiny primary accent color reminiscent of pure
# and clear ice.
nord8 = "h110"

# nord9 is a more darkened and less saturated color reminiscent of
# arctic waters.
nord9 = "h111"

# nord10 is a dark and intensive color reminiscent of the deep arctic ocean.
nord10 = "h67"

# ------------------------------------------------------------------------------
# Aurora consists of five colorful components reminiscent of the "Aurora
# borealis", sometimes referred to as polar lights or northern lights.
# ------------------------------------------------------------------------------

# nord11 is a reddish color.
nord11 = "h138"

# nord12 is an orangey color.
nord12 = "h174"

# nord13 is a yellowy color.
nord13 = "h216"

# nord14 is a greenish color.
nord14 = "h150"

# nord15 is a purplish color.
nord15 = "h139"

link("current breakpoint", "current frame name")
link("focused current breakpoint", "focused current frame name")
palette_dict = {

    # {{{ base styles
    "background": (nord0, nord7),
    "selectable": (nord4, nord1),
    "focused selectable": (nord4, nord2),
    "highlighted": (nord14, nord1),
    "hotkey": (add_setting(nord0, "underline"), nord7),
    # }}}
    # {{{ general ui
    "input": (nord4, nord0),
    "button": (add_setting(nord4, "bold"), nord0),
    "focused button": (add_setting(nord4, "bold"), nord2),
    "focused sidebar": (nord0, nord7),
    "warning": (nord12, nord2),
    "group head": (add_setting(nord0, "bold"), nord7),
    "dialog title": (add_setting(nord4, "bold"), nord1),
    # }}}
    # {{{ source view
    "source": (nord4, nord0),
    "current source": (nord13, nord1),
    "current focused source": (nord13, nord2),
    "breakpoint source": (nord12, nord0),
    "line number": (nord2, nord0),
    "current line marker": (add_setting(nord13, "bold"), nord0),
    "breakpoint marker": (add_setting(nord13, "bold"), nord0),
    # }}}
    # {{{ sidebar
    "sidebar two": (nord7, nord1),
    "focused sidebar two": (nord7, nord2),
    "sidebar three": (nord10, nord1),
    "focused sidebar three": (nord10, nord2),
    # }}}
    # {{{ variables view
    "highlighted var label": (nord14, nord1),
    "focused highlighted var label": (nord14, nord2),
    "focused highlighted var value": (nord14, nord2),
    "return label": (nord0, nord8),
    "focused return label": (nord0, nord8),
    # }}}
    # {{{ stack
    "current frame name": (nord14, nord1),
    "focused current frame name": (nord14, nord2),
    # }}}
    # {{{ shell
    "command line prompt": (add_setting(nord5, "bold"), nord0),
    "command line output": (nord7, nord0),
    "command line error": (nord12, nord0),
    "focused command line output": (nord7, nord2),
    "focused command line error": (nord12, nord2),
    # }}}
    # {{{ Code syntax
    "literal":     (nord15, nord0),
    "builtin":     (add_setting(nord7, "bold"), nord0),
    "pseudo":      (nord7, nord0),
    "exception":   (nord11, nord0),
    "function":    (nord7, nord0),
    "class":       (add_setting(nord9, "underline"), nord0),
    "keyword":     (nord8, nord0),
    "keyword2":    (nord9, nord0),
    "operator":    (nord8, nord0),
    "comment":     (nord3, nord0),
    "string":      (nord14, nord0),
    # }}}
}

# vim: foldmethod=marker



from pudb.themes.utils import add_setting


palette_dict = {
    # {{{ base styles
    "background": ("light green", "light gray"),
    "selectable": ("light green", "white"),
    "focused selectable": ("white", "dark blue"),
    "highlighted": ("white", "dark cyan"),
    "hotkey": (add_setting("black", "underline"), "light gray"),
    # }}}
    # {{{ general ui
    "dialog title": (add_setting("white", "bold"), "dark cyan"),
    "warning": (add_setting("light red", "bold"), "white"),
    "header warning": (add_setting("light red", "bold"), "light gray"),
    "focused sidebar": ("dark red", "light gray"),
    "group head": (add_setting("yellow", "bold"), "light gray"),
    # }}}
    # {{{ source view
    "source": ("yellow", "white"),
    "breakpoint source": ("light red", "light gray"),
    "current source": ("light gray", "light blue"),
    "line number": ("light blue", "white"),
    "current line marker": (add_setting("light blue", "bold"), "white"),
    "breakpoint marker": (add_setting("light red", "bold"), "white"),
    # }}}
    # {{{ sidebar
    "sidebar two": ("dark blue", "white"),
    "sidebar three": ("light cyan", "white"),
    "focused sidebar three": ("light gray", "dark blue"),
    # }}}
    # {{{ variables view
    "return label": ("white", "yellow"),
    "focused return label": ("white", "yellow"),
    # }}}
    # {{{ stack
    "current frame name": (add_setting("light green", "bold"), "white"),
    "focused current frame name": (add_setting("white", "bold"), "dark blue"),
    # }}}
    # {{{ shell
    "command line output": ("light green", "white"),
    # }}}
    # {{{ Code syntax
    "namespace": ("dark red", "white"),
    "exception": ("light red", "white"),
    "keyword": ("brown", "white"),
    "keyword2": ("dark magenta", "white"),
    "function": ("dark green", "white"),
    "literal": ("dark cyan", "white"),
    "builtin": ("dark blue", "white"),
    "comment": ("light cyan", "white"),
    "pseudo": ("light cyan", "white"),
    # }}}
}

# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = """
Copyright (C) 2009-2017 Andreas Kloeckner
Copyright (C) 2014-2017 Aaron Meurer
"""

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from dataclasses import astuple, dataclass, replace

from pudb.lowlevel import ui_log
from pudb.themes import THEMES
from pudb.themes.utils import (
    add_setting,
    inheritance_overrides,
    link,
    reset_inheritance_overrides,
)


@dataclass
class PaletteEntry:
    name: str
    foreground: str = "default"
    background: str = "default"
    mono: str | None = None
    foreground_high: str | None = None
    background_high: str | None = None

    def handle_256_colors(self):
        if self.foreground.lower().strip().startswith("h"):
            self.foreground_high = self.foreground
            self.foreground = "default"
        if self.background.lower().strip().startswith("h"):
            self.background_high = self.background
            self.background = "default"


# ------------------------------------------------------------------------------
# Reference for some palette items:
#
#  "namespace" : "import", "from", "using"
#  "operator"  : "+", "-", "=" etc.
#                NOTE: Does not include ".", which is assigned the type "source"
#  "argument"  : Function arguments
#  "builtin"   : "range", "dict", "set", "list", etc.
#  "pseudo"    : "self", "cls"
#  "dunder"    : Class method names of the form __<name>__ within
#               a class definition
#  "magic"     : Subset of "dunder", methods that the python language assigns
#               special meaning to. ("__str__", "__init__", etc.)
#  "exception" : Exception names
#  "keyword"   : All keywords except those specifically assigned to "keyword2"
#                ("from", "and", "break", "is", "try", "True", "None", etc.)
#  "keyword2"  : "class", "def", "exec", "lambda", "print"
# ------------------------------------------------------------------------------


# {{{ style inheritance
BASE_STYLES = {
    "background": None,
    "selectable": None,
    "focused selectable": None,
    "highlighted": None,
    "hotkey": None,
}


# Map styles to their parent. If a style is not defined, use the parent style
# recursively.
# focused > highlighted > current > breakpoint line/disabled breakpoint
INHERITANCE_MAP = {
    # {{{ general ui
    "label": "background",
    "header": "background",
    "dialog title": "header",
    "group head": "header",
    "focused sidebar": "header",

    "input": "selectable",
    "focused input": "focused selectable",
    "button": "input",
    "focused button": "focused input",
    "value": "input",
    "fixed value": "label",

    "warning": "highlighted",
    "header warning": "warning",
    "search box": "focused input",
    "search not found": "warning",
    # }}}

    # {{{ source view
    "source": "selectable",
    "focused source": "focused selectable",
    "highlighted source": "highlighted",

    "current source": "source",
    "current focused source": "focused source",
    "current highlighted source": "current source",

    "breakpoint source": "source",
    "breakpoint focused source": "focused source",
    "current breakpoint source": "current source",
    "current breakpoint focused source": "current focused source",

    "line number": "source",
    "breakpoint marker": "line number",
    "current line marker": "breakpoint marker",
    # }}}

    # {{{ sidebar
    "sidebar one": "selectable",
    "sidebar two": "selectable",
    "sidebar three": "selectable",

    "focused sidebar one": "focused selectable",
    "focused sidebar two": "focused selectable",
    "focused sidebar three": "focused selectable",
    # }}}

    # {{{ variables view
    "variables": "selectable",
    "variable separator": "background",

    "var value": "sidebar one",
    "var label": "sidebar two",
    "focused var value": "focused sidebar one",
    "focused var label": "focused sidebar two",

    "highlighted var label": "highlighted",
    "highlighted var value": "highlighted",
    "focused highlighted var label": "focused var label",
    "focused highlighted var value": "focused var value",

    "return label": "var label",
    "return value": "var value",
    "focused return label": "focused var label",
    "focused return value": "focused var value",
    # }}}

    # {{{ stack
    "stack": "selectable",

    "frame name": "sidebar one",
    "frame class": "sidebar two",
    "frame location": "sidebar three",

    "focused frame name": "focused sidebar one",
    "focused frame class": "focused sidebar two",
    "focused frame location": "focused sidebar three",

    "current frame name": "frame name",
    "current frame class": "frame class",
    "current frame location": "frame location",

    "focused current frame name": "focused frame name",
    "focused current frame class": "focused frame class",
    "focused current frame location": "focused frame location",
    # }}}

    # {{{ breakpoints view
    "breakpoint": "sidebar two",
    "disabled breakpoint": "sidebar three",
    "current breakpoint": "breakpoint",
    "disabled current breakpoint": "disabled breakpoint",

    "focused breakpoint": "focused sidebar two",
    "focused current breakpoint": "focused breakpoint",
    "focused disabled breakpoint": "focused sidebar three",
    "focused disabled current breakpoint": "focused disabled breakpoint",
    # }}}

    # {{{ shell
    "command line edit": "source",
    "command line output": "source",
    "command line prompt": "source",
    "command line input": "source",
    "command line error": "warning",

    "focused command line output": "focused source",
    "focused command line input": "focused source",
    "focused command line error": "focused source",

    "command line clear button": "button",
    "command line focused button": "focused button",
    # }}}

    # {{{ Code syntax
    "comment":      "source",
    "keyword":      "source",
    "literal":      "source",
    "name":         "source",
    "operator":     "source",
    "punctuation":  "source",
    "argument":     "name",
    "builtin":      "name",
    "exception":    "name",
    "function":     "name",
    "pseudo":       "builtin",
    "class":        "function",
    "dunder":       "function",
    "magic":        "dunder",
    "namespace":    "keyword",
    "keyword2":     "keyword",
    "string":       "literal",
    "doublestring": "string",
    "singlestring": "string",
    "docstring":    "string",
    "backtick":     "string",
    # }}}
}


def set_style(palette_dict: dict, style_name: str,
              inheritance_overrides: dict) -> PaletteEntry:
    """
    Recursively search up the style hierarchy for the first style which has
    been defined, and add it to the palette_dict under the given style_name.
    """
    try:
        style = palette_dict[style_name]
        if not isinstance(style, PaletteEntry):
            style = PaletteEntry(style_name, *style)
            style.handle_256_colors()
            palette_dict[style_name] = style
        return style
    except KeyError:
        parent_name = inheritance_overrides.get(
            style_name,
            INHERITANCE_MAP[style_name],
        )
        style = replace(
            set_style(palette_dict, parent_name, inheritance_overrides),
            name=style_name
        )
        palette_dict[style_name] = style
        return style

# }}}


# {{{ get palette

def get_palette(may_use_fancy_formats: bool, theme: str = "classic") -> list:
    """
    Load the requested theme and return a list containing all palette entries
    needed to highlight the debugger UI, including syntax highlighting.
    """
    reset_inheritance_overrides()

    try:
        palette_dict = THEMES[theme]
    except KeyError:
        # {{{ custom
        try:
            # {{{ base styles
            palette_dict = {
                "background": ("black", "light gray"),
                "hotkey": (add_setting("black", "underline"), "light gray"),
                "selectable": ("black", "dark cyan"),
                "focused selectable": ("black", "dark green"),
                "input": (add_setting("yellow", "bold"), "dark blue"),
                "warning": (add_setting("white", "bold"), "dark red"),
                "highlighted": ("white", "dark cyan"),
                "source": ("white", "dark blue"),
            }
            # }}}

            symbols = {
                "palette": palette_dict,
                "add_setting": add_setting,
                "link": link,
            }

            from os.path import expanduser, expandvars
            fname = expanduser(expandvars(theme))
            with open(fname) as inf:
                exec(compile(inf.read(), fname, "exec"), symbols)
        except FileNotFoundError:
            ui_log.error(f"Unable to locate custom theme file {theme!r}"
                         )
            return None
        except Exception:
            ui_log.exception("Error when importing theme:")
            return None
        # }}}

    # Apply style inheritance
    for style_name in set(INHERITANCE_MAP.keys()).union(BASE_STYLES.keys()):
        set_style(palette_dict, style_name, inheritance_overrides)

    palette_list = [
        astuple(entry)
        for entry in palette_dict.values()
        if isinstance(entry, PaletteEntry)
    ]

    return palette_list

# }}}

# vim: foldmethod=marker



from pudb.themes.utils import add_setting, link


link("current breakpoint", "current frame name")
link("focused current breakpoint", "focused current frame name")

palette_dict = {
    # {{{ base styles
    "background": ("black", "light gray"),
    "selectable": ("black", "dark cyan"),
    "focused selectable": ("black", "light cyan"),
    "hotkey": (add_setting("black", "bold, underline"), "light gray"),
    "highlighted": ("black", "yellow"),
    # }}}
    # {{{ general ui
    "header": (add_setting("black", "bold"), "light gray"),
    "group head": ("dark blue", "light gray"),
    "dialog title": (add_setting("white", "bold"), "dark blue"),
    "input": ("black", "dark cyan"),
    "focused input": ("black", "light cyan"),
    "warning": (add_setting("dark red", "bold"), "white"),
    "header warning": (add_setting("dark red", "bold"), "light gray"),
    # }}}
    # {{{ source view
    "source": ("black", "white"),
    "current source": ("black", "dark cyan"),
    "breakpoint source": ("dark red", "light gray"),
    "line number": ("dark gray", "white"),
    "current line marker": ("dark red", "white"),
    "breakpoint marker": ("dark red", "white"),
    # }}}
    # {{{ sidebar
    "sidebar one": ("black", "dark cyan"),
    "sidebar two": ("dark blue", "dark cyan"),
    "sidebar three": ("dark gray", "dark cyan"),
    "focused sidebar one": ("black", "light cyan"),
    "focused sidebar two": ("dark blue", "light cyan"),
    "focused sidebar three": ("dark gray", "light cyan"),
    # }}}
    # {{{ variables view
    "highlighted var label": ("dark blue", "yellow"),
    "return label": ("white", "dark blue"),
    "focused return label": ("light gray", "dark blue"),
    # }}}
    # {{{ stack
    "current frame name": (add_setting("white", "bold"), "dark cyan"),
    "focused current frame name": (add_setting("black", "bold"), "light cyan"),
    # }}}
    # {{{ shell
    "command line output": (
        add_setting("dark gray", "bold"), "white"),
    # }}}
    # {{{ Code syntax
    "keyword2": ("dark magenta", "white"),
    "namespace": ("dark magenta", "white"),
    "literal": ("dark red", "white"),
    "exception": ("dark red", "white"),
    "comment": ("dark gray", "white"),
    "function": ("dark blue", "white"),
    "pseudo": ("dark gray", "white"),
    "builtin": ("light blue", "white"),
    # }}}
}

# vim: foldmethod=marker



from .agr_256 import palette_dict as agr256
from .classic import palette_dict as classic
from .dark_vim import palette_dict as darkvim
from .gray_light_256 import palette_dict as graylight256
from .midnight import palette_dict as midnight
from .mono import palette_dict as mono
from .monokai import palette_dict as monokai
from .monokai_256 import palette_dict as monokai256
from .nord_dark_256 import palette_dict as norddark256
from .solarized import palette_dict as solarized
from .vim import palette_dict as vim


THEMES = {
    "classic": classic,
    "vim": vim,
    "dark vim": darkvim,
    "midnight": midnight,
    "monokai": monokai,
    "solarized": solarized,
    "mono": mono,
    "agr-256": agr256,
    "gray-light-256": graylight256,
    "monokai-256": monokai256,
    "nord-dark-256": norddark256,
}



from pudb.themes.utils import add_setting, link


link("current breakpoint", "current frame name")
link("focused current breakpoint", "focused current frame name")

palette_dict = {
    # {{{ base styles
    "background": ("black", "light gray"),
    "selectable": ("white", "dark gray"),
    "focused selectable": (add_setting("white", "bold"), "light blue"),
    "highlighted": ("black", "dark green"),
    "hotkey": (add_setting("dark blue", "underline"), "light gray"),
    # }}}
    # {{{ general ui
    "header": ("dark blue", "light gray"),
    "dialog title": (add_setting("white", "bold"), "black"),
    "warning": (add_setting("light red", "bold"), "black"),
    "header warning": (add_setting("light red", "bold"), "light gray"),
    # }}}
    # {{{ source view
    "source": ("white", "black"),
    "current source": (add_setting("white", "bold"), "dark gray"),
    "line number": (add_setting("dark gray", "bold"), "black"),
    "breakpoint marker": (add_setting("light red", "bold"), "black"),
    "breakpoint source": (add_setting("white", "bold"), "dark red"),
    # }}}
    # {{{ sidebar
    "sidebar two": ("yellow", "dark gray"),
    "focused sidebar two": ("light cyan", "light blue"),
    "sidebar three": ("light gray", "dark gray"),
    "focused sidebar three": ("yellow", "light blue"),
    # }}}
    # {{{ stack
    "current frame name": (
        add_setting("white", "bold"), "dark gray"),
    # }}}
    # {{{ shell
    "command line output": (add_setting("yellow", "bold"), "black"),
    # }}}
    # {{{ Code syntax
    "keyword": ("yellow", "black"),
    "literal": ("light magenta", "black"),
    "function": (add_setting("light cyan", "bold"), "black"),
    "punctuation": ("yellow", "black"),
    "comment": ("dark cyan", "black"),
    "exception": ("light red", "black"),
    "builtin": ("light green", "black"),
    "pseudo": ("dark green", "black"),
    # }}}
}

# vim: foldmethod=marker



from __future__ import annotations

import os
import sys

from pudb.debugger import Debugger


def set_trace(paused=True, frame=None, term_size=None):
    """Set a breakpoint in a forked process on Unix system, e.g. Linux & MacOS.
    In- and output will be redirected to /dev/stdin & /dev/stdout.
    You can call pudb.forked.set_trace() directly or
    use it with python's built-in breakpoint():
    PYTHONBREAKPOINT=pudb.forked.set_trace python …
    """
    if frame is None:
        frame = sys._getframe().f_back
    if term_size is None:
        term_size = os.environ.get("PUDB_TERM_SIZE")
        if term_size is not None:
            term_size = tuple(map(int, term_size.split("x")))
            if len(term_size) != 2:
                raise ValueError("PUDB_TERM_SIZE should have two dimensions")
        else:
            try:
                # Getting terminal size
                s = os.get_terminal_size()
                term_size = (s.columns, s.lines)
            except Exception:
                term_size = (80, 24)

    Debugger(
        stdin=open("/dev/stdin"),
        stdout=open("/dev/stdout", "w"),
        term_size=term_size,
    ).set_trace(frame, paused=paused)



from pudb.themes.utils import add_setting


palette_dict = {
    # {{{ base styles
    "background": ("h232", "h248"),
    "selectable": ("h232", "h252"),
    "focused selectable": ("h232", "h251"),
    "highlighted": (add_setting("h234", "bold, underline"), "h252"),
    "hotkey": (add_setting("h232", "underline"), "h248"),
    # }}}
    # {{{ general ui
    "focused sidebar": (add_setting("h232", "bold"), "h248"),
    "warning": (add_setting("h232", "bold"), "h253"),
    "group head": (add_setting("h232", "bold"), "h248"),
    "dialog title": (add_setting("h232", "underline, bold"), "h248"),
    # }}}
    # {{{ source view
    "source": ("h235", "h253"),
    "current source": (add_setting("h232", "underline"), "h253"),
    "line number": ("h244", "h253"),
    # }}}
    # {{{ sidebar
    "sidebar two": (add_setting("h234", "bold"), "h252"),
    "focused sidebar two": (add_setting("h234", "bold"), "h251"),
    "sidebar three": ("h239", "h252"),
    "focused sidebar three": ("h239", "h251"),
    # }}}
    # {{{ Code syntax
    "exception":   (add_setting("h236", "underline"), "h253"),
    "class":       (add_setting("h234", "bold, underline"), "h253"),
    "keyword":     (add_setting("h234", "bold"), "h253"),
    "comment":     ("h244", "h253"),
    # }}}
}



from __future__ import annotations


__copyright__ = """
Copyright (C) 2009-2017 Andreas Kloeckner
Copyright (C) 2014-2017 Aaron Meurer
Copyright (C) 2024 Gerhard Sittig
"""

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


import logging
import sys
from datetime import datetime
from enum import Enum, auto
from typing import TYPE_CHECKING, TypeVar, cast

from typing_extensions import override


if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Sequence
    from types import CodeType


logfile: list[str | None] = [None]


def getlogfile():
    return logfile[0]


def setlogfile(destfile: str):
    logfile[0] = destfile
    with open(destfile, "a") as openfile:
        openfile.write(
            f"\n*** Pudb session error log started at {datetime.now()} ***\n")


# once we've dropped Py3.8 support, use this instead:
# class TerminalOrStreamHandler(logging.StreamHandler[TextIOWrapper]):
class TerminalOrStreamHandler(logging.StreamHandler):
    """
    Logging handler that sends errors either to the terminal window or to
    stderr, depending on whether the debugger is active.
    """
    @override
    def emit(self, record: logging.LogRecord):
        from pudb import _get_debugger, _have_debugger
        logfile = getlogfile()

        self.acquire()
        try:
            if logfile is not None:
                message = self.format(record)
                with open(logfile, "a") as openfile:
                    openfile.write(f"\n{message}\n")
            elif _have_debugger():
                dbg = _get_debugger()
                message = self.format(record)
                dbg.ui.add_cmdline_content(message, "command line error")
            else:
                super().emit(record)
        finally:
            self.release()


def _init_loggers():
    ui_handler = TerminalOrStreamHandler()
    ui_formatter = logging.Formatter(
        fmt="*** Pudb UI Exception Encountered: %(message)s ***\n"
    )
    ui_handler.setFormatter(ui_formatter)
    ui_log = logging.getLogger("ui")
    ui_log.addHandler(ui_handler)

    settings_handler = TerminalOrStreamHandler()
    settings_formatter = logging.Formatter(
        fmt="*** Pudb Settings Exception Encountered: %(message)s ***\n"
    )
    settings_handler.setFormatter(settings_formatter)
    settings_log = logging.getLogger("settings")
    settings_log.addHandler(settings_handler)

    return ui_log, settings_log


ui_log, settings_log = _init_loggers()


# {{{ breakpoint validity

def generate_executable_lines_for_code(code: CodeType) -> Iterable[int]:
    if sys.version_info >= (3, 10):
        for _start, _end, lineno in code.co_lines():
            if lineno is not None:
                yield lineno
    else:
        lineno = code.co_firstlineno
        yield lineno
        # See https://github.com/python/cpython/blob/master/Objects/lnotab_notes.txt

        for line_incr in code.co_lnotab[1::2]:
            # NB: This code is specific to Python 3.6 and higher
            # https://github.com/python/cpython/blob/v3.6.0/Objects/lnotab_notes.txt
            if line_incr >= 0x80:
                line_incr -= 0x100
            lineno += line_incr
            yield lineno


def get_executable_lines_for_codes_recursive(codes: Sequence[CodeType]):
    codes = list(codes)

    from types import CodeType

    execable_lines: set[int] = set()

    while codes:
        code = codes.pop()
        execable_lines |= set(generate_executable_lines_for_code(code))
        codes.extend(const
                for const in code.co_consts  # pyright: ignore[reportAny]
                if isinstance(const, CodeType))

    return execable_lines


def get_executable_lines_for_file(filename: str):
    # inspired by rpdb2

    from linecache import getlines
    codes = [compile("".join(getlines(filename)), filename, "exec", dont_inherit=1)]

    return get_executable_lines_for_codes_recursive(codes)


def get_breakpoint_invalid_reason(filename: str, lineno: int):
    # simple logic stolen from pdb
    import linecache
    line = linecache.getline(filename, lineno)
    if not line:
        return "Line is beyond end of file."

    try:
        executable_lines = get_executable_lines_for_file(filename)
    except SyntaxError:
        return "File failed to compile."

    if lineno not in executable_lines:
        return "No executable statement found in line."


def lookup_module(filename: str):
    """Helper function for break/clear parsing -- may be overridden.

    lookupmodule() translates (possibly incomplete) file or module name
    into an absolute file name.
    """

    # stolen from pdb
    import os
    import sys

    if os.path.isabs(filename) and os.path.exists(filename):
        return filename
    f = os.path.join(sys.path[0], filename)
    if os.path.exists(f):  # and self.canonic(f) == self.mainpyfile:
        return f
    _root, ext = os.path.splitext(filename)
    if ext == "":
        filename = filename + ".py"
    if os.path.isabs(filename):
        return filename
    for dirname in sys.path:
        while os.path.islink(dirname):
            dirname = os.readlink(dirname)
        fullname = os.path.join(dirname, filename)
        if os.path.exists(fullname):
            return fullname
    return None

# }}}


# {{{ file encoding detection
# the main idea stolen from Python 3.1's tokenize.py, by Ka-Ping Yee

import re


cookie_re = re.compile(br"^\s*#.*coding[:=]\s*([-\w.]+)")
from codecs import BOM_UTF8, lookup


StrOrBytesT = TypeVar("StrOrBytesT", str, bytes, "str | bytes")


def detect_encoding(
            line_iter: Iterator[StrOrBytesT]
        ) -> tuple[str | None, Sequence[StrOrBytesT]]:
    """
    The detect_encoding() function is used to detect the encoding that should
    be used to decode a Python source file. It requires one argument, line_iter,
    an iterator on the lines to be read.

    It will read a maximum of two lines, and return the encoding used
    (as a string) and a list of any lines (left as bytes) it has read
    in.

    It detects the encoding from the presence of a utf-8 bom or an encoding
    cookie as specified in pep-0263. If both a bom and a cookie are present,
    but disagree, a SyntaxError will be raised. If the encoding cookie is an
    invalid charset, raise a SyntaxError.

    If no encoding is specified, then the default of 'utf-8' will be returned.
    """
    bom_found = False

    def read_or_stop():
        try:
            return next(line_iter)
        except StopIteration:
            return ""

    def find_cookie(line: bytes):
        try:
            line_string = line
        except UnicodeDecodeError:
            return None

        matches = cookie_re.findall(line_string)
        if not matches:
            return None
        encoding = matches[0].decode()
        try:
            codec = lookup(encoding)
        except LookupError as err:
            # This behaviour mimics the Python interpreter
            raise SyntaxError("unknown encoding: " + encoding) from err

        if bom_found and codec.name != "utf-8":
            # This behaviour mimics the Python interpreter
            raise SyntaxError("encoding problem: utf-8")
        return encoding

    first = read_or_stop()
    if isinstance(first, str):
        return None, [first]

    if first.startswith(BOM_UTF8):
        bom_found = True
        first = first[3:]
    if not first:
        return "utf-8", []

    encoding = find_cookie(first)
    if encoding:
        return encoding, [first]

    second = read_or_stop()
    if not second:
        return "utf-8", [first]

    encoding = find_cookie(cast("bytes", second))
    if encoding:
        return encoding, [first, second]

    return "utf-8", [first, second]


def decode_lines(lines: Iterable[bytes | str]) -> Iterable[str]:
    line_iter = iter(lines)
    source_enc, detection_read_lines = detect_encoding(line_iter)

    from itertools import chain

    for line in chain(detection_read_lines, line_iter):
        if hasattr(line, "decode") and source_enc is not None:
            yield line.decode(source_enc)
        else:
            yield cast("str", line)

# }}}


# {{{ get single key press from console outside of curses

class KeyReadImpl(Enum):
    INPUT = auto()
    GETCH = auto()
    SELECT = auto()


_keyread_impl = KeyReadImpl.INPUT
if sys.platform in ("emscripten", "wasi"):
    pass
elif sys.platform in ("win32",):
    _keyread_impl = KeyReadImpl.GETCH
else:
    _keyread_impl = KeyReadImpl.SELECT


class ConsoleSingleKeyReader:
    """
    Get a single key press from a terminal without a prompt.

    Eliminates the necessity to press ENTER before other input also
    becomes available. Avoids the accumulation of prompts on the screen
    as was the case with Python's input() call. Is used in situations
    where urwid is disabled and curses calls are not available.

    Supports major desktop platforms with special cases (msvcrt getch(),
    termios and select). Transparently falls back to Python's input()
    method. Call sites remain simple and straight forward.
    """

    def __enter__(self):
        if _keyread_impl == KeyReadImpl.SELECT:
            import termios
            import tty
            self.prev_settings = termios.tcgetattr(sys.stdin)
            tty.setcbreak(sys.stdin.fileno())
        return self

    def __exit__(self, type, value, traceback):
        if _keyread_impl == KeyReadImpl.SELECT:
            import termios
            termios.tcsetattr(sys.stdin, termios.TCSADRAIN, self.prev_settings)

    def get_single_key(self):
        if _keyread_impl == KeyReadImpl.GETCH:
            import msvcrt
            # https://docs.python.org/3/library/msvcrt.html#msvcrt.getch
            # Most keys are returned in the first getch() call. Some
            # special keys (function keys, cursor, keypad) require
            # another call when the first returned '\0' or '\xe0'.
            c = msvcrt.getch()
            if c in ("\x00", "\xe0"):
                c = msvcrt.getch()
            return c

        elif _keyread_impl == KeyReadImpl.SELECT:
            import select
            rset, _, _ = select.select([sys.stdin], [], [], None)
            assert sys.stdin in rset
            return sys.stdin.read(1)

        # Strictly speaking putting the fallback here which requires
        # pressing ENTER is not correct, this is the "non buffered"
        # console support code. But it simplifies call sites. And is
        # easy to tell by users because a prompt is provided. This is
        # the most portable approach, and backwards compatible with
        # earlier PuDB releases. It's a most appropriate default for
        # otherwise unsupported platforms. Or when users choose to
        # not accept single key presses, or keys other than ENTER.
        else:
            input("Hit Enter to return:")
            return None

# }}}

# vim: foldmethod=marker



from pudb.themes.utils import add_setting, link


# Based on XCode's midnight theme
# Looks best in a console with green text against black background
link("current breakpoint", "current frame name")
link("focused current breakpoint", "focused current frame name")

palette_dict = {
    # {{{ base styles
    "background": ("black", "light gray"),
    "selectable": ("white", "black"),
    "focused selectable": ("white", "dark blue"),
    "hotkey": (add_setting("black", "underline, italics"), "light gray"),
    "highlighted": ("black", "dark green"),
    # }}}
    # {{{ general ui
    "input": ("light green", "black"),
    "focused input": ("light green", "black"),
    "warning": (add_setting("white", "bold"), "dark red"),
    "dialog title": (add_setting("white", "bold"), "dark blue"),
    "group head": (add_setting("dark blue", "bold"), "light gray"),
    "button": (add_setting("white", "bold"), "dark blue"),
    "focused button": ("white", "black"),
    "focused sidebar": ("black", "white"),
    "value": (add_setting("yellow", "bold"), "dark blue"),
    # }}}
    # {{{ source view
    "source": ("light green", "black"),
    "highlighted source": ("black", "dark green"),
    "current source": ("black", "brown"),
    "current focused source": (add_setting("yellow", "bold"), "dark blue"),
    "breakpoint source": (add_setting("yellow", "bold"), "dark red"),
    "current breakpoint source": ("black", "dark red"),

    "line number": ("light gray", "black"),
    "current line marker": ("dark red", "black"),
    "breakpoint marker": ("dark red", "black"),
    # }}}
    # {{{ sidebar
    "sidebar two": ("light blue", "black"),
    "sidebar three": ("light cyan", "black"),
    # }}}
    # {{{ variables view
    "return label": ("white", "dark blue"),
    "return value": ("black", "dark cyan"),
    "focused return label": ("light gray", "dark blue"),
    # }}}
    # {{{ stack
    "current frame name": (add_setting("white", "bold"), "black"),
    "current frame class": (add_setting("light blue", "bold"), "black"),
    "current frame location": (add_setting("light cyan", "bold"), "black"),

    "focused current frame name": (add_setting("white", "bold"), "dark blue"),
    "focused current frame class": (add_setting("white", "bold"), "dark blue"),
    "focused current frame location": (add_setting("white", "bold"), "dark blue"),
    # }}}
    # {{{ breakpoints view
    "breakpoint": ("white", "black"),
    "disabled breakpoint": ("dark gray", "black"),
    "focused disabled breakpoint": ("light gray", "dark blue"),
    "current breakpoint": (add_setting("white", "bold"), "black"),
    "disabled current breakpoint": (add_setting("dark gray", "bold"), "black"),
    "focused current breakpoint": (add_setting("white", "bold"), "dark blue"),
    "focused disabled current breakpoint": (
        add_setting("light gray", "bold"), "dark blue"),
    # }}}
    # {{{ shell
    "command line edit": ("white", "black"),
    "command line prompt": (add_setting("white", "bold"), "black"),

    "command line input": ("white", "black"),
    "command line error": (add_setting("light red", "bold"), "black"),

    "command line clear button": (add_setting("white", "bold"), "black"),
    "command line focused button": ("white", "dark blue"),
    # }}}
    # {{{ Code syntax
    "keyword": ("dark magenta", "black"),
    "operator": ("dark green", "black"),
    "pseudo": ("light magenta", "black"),
    "function": (add_setting("light blue", "bold"), "black"),
    "builtin": ("dark gray", "black"),
    "literal": ("dark cyan", "black"),
    "string": ("dark red", "black"),
    "docstring": ("yellow", "black"),
    "backtick": ("dark green", "black"),
    "punctuation": ("white", "black"),
    "comment": ("white", "black"),
    "exception": ("dark green", "black"),
    # }}}
}

# vim: foldmethod=marker
